(***********************************************************************)
(*                                                                     *)
(*                             Pio (OCaml)                             *)
(*                                                                     *)
(*                    Polyglot Input/Output Library                    *)
(*                                                                     *)
(*               EPI Pomdapi, INRIA Paris - Rocquencourt               *)
(*                                                                     *)
(*  Copyright 2005-2014 INRIA.                                         *)
(*  Institut National de Recherche en Informatique et en Automatique.  *)
(*  All rights reserved. This file is distributed only by permission.  *)
(*                                                                     *)
(*  Francois Clement <Francois.Clement@inria.fr>                       *)
(*  Pierre Weis <Pierre.Weis@inria.fr>                                 *)
(*                                                                     *)
(***********************************************************************)

(* $Id: pio.mli,v 1.28 2014/03/21 10:11:01 fclement Exp $ *)

(** {3 The Caml API for [PIO]}
 
 We describe the communication library between Caml programs and foreign
 programs. *)

(** {6 Naming conventions}
 
 The Caml [PIO] library provides input/output primitives operating on channels.
 The naming convention is:
 - primitives for reading a <type> from a given input channel are named
   [input_]<type>,
 - primitives for writing a <type> into a given output channel are named
   [output_]<type>.
 
 The Caml [PIO] library provides input/output primitives operating on buffers.
 The naming convention is:
 - primitives for reading a <type> from a given input buffer
   (a [Scanf.Scanning.scanbuf] value) are named [read_]<type>,
 - primitives for writing a <type> into a given output buffer
   (a [Buffer.t] value) are named [write_]<type>.
 
 In addition, the [PIO] library provides acknowledged input/output
 primitives operating on buffers. Acknowledged input/output primitives
 involve an acknowledgment mechanism to ensure that the communication has
 been properly delivered (indeed read or indeed written).
 The naming convention is:
 - primitives for acknowledged reading a <type> from a given input buffer
   (a [Scanf.Scanning.scanbuf] value) are named [receive_]<type>,
 - primitives for acknowledged writing a <type> into a given output buffer
   (a [Buffer.t] value) are named [send_]<type>.
 
 For the time being, the acknowledged level is not yet implemented and the
 [receive_] and [send_] primitives are equivalent to their [read_] and
 [write_] counterparts. *)

(** {6 Output communication functions} *)

(** {7 Functions to send communications}
 
 Those functions send a communication to a buffer.
 
 More precisely, they write a {!Communication.t} value on a Buffer.t
 output buffer. *)

val send_one_value : Buffer.t -> Typed_value.t -> unit;;
(** [send_one_value ob v] sends to the Buffer.t output buffer [ob] a
 [Phrase] communication containing the unique {!Typed_value.t} value [v]. *)

val send_values : Buffer.t -> Typed_value.t array -> unit;;
(** [send_values ob vals] sends to the Buffer.t output buffer [ob] a
 [Phrase] communication containing all the typed values of the vector of
 {!Typed_value.t} values [vals]. *)

val send_communication : Buffer.t -> Communication.t -> unit;;
(** [send_communication ob com] sends its {!Communication.t} [comm] argument
 to the Buffer.t output buffer [ob]. *)

(** {7 Functions to send typed values}
 
 Those functions send a regular Caml value to a buffer.
 
 More precisely, they write the {!Typed_value.t} value corresponding to their
 Caml value argument on a Buffer.t output buffer. *)

val send_string : Buffer.t -> string -> unit;;
(** [send_string ob s] sends the string [s] as a {!Typed_value.t} typed
 value to the Buffer.t output buffer [ob]. *)

val send_int : Buffer.t -> int -> unit;;
(** Same as above, for an integer. *)

val send_int_couple : Buffer.t -> int * int -> unit;;
(** Same as above, for an integer couple. *)

val send_int_triple : Buffer.t -> int * int * int -> unit;;
(** Same as above, for an integer triple. *)

val send_float : Buffer.t -> float -> unit;;
(** Same as above, for a float. *)

val send_float_couple : Buffer.t -> float * float -> unit;;
(** Same as above, for a float couple. *)

val send_float_triple : Buffer.t -> float * float * float -> unit;;
(** Same as above, for a float triple. *)

val send_string_vector : Buffer.t -> string array -> unit;;
(** Same as above, for a string array. *)

val send_int_vector : Buffer.t -> int array -> unit;;
(** Same as above, for an integer array. *)

val send_int_couple_vector : Buffer.t -> (int * int) array -> unit;;
(** Same as above, for an integer couple array. *)

val send_int_triple_vector : Buffer.t -> (int * int * int) array -> unit;;
(** Same as above, for an integer triple array. *)

val send_float_vector : Buffer.t -> float array -> unit;;
(** Same as above, for a float array. *)

val send_float_couple_vector : Buffer.t -> (float * float) array -> unit;;
(** Same as above, for a float couple array. *)

val send_float_triple_vector :
  Buffer.t -> (float * float * float) array -> unit
;;
(** Same as above, for a float triple array. *)

val send_c_int_matrix : Buffer.t -> int Umatrix.t -> unit;;
(** Same as above, for a C integer matrix. *)

val send_fortran_int_matrix : Buffer.t -> int Umatrix.t -> unit;;
(** Same as above, for a Fortran integer matrix. *)

val send_c_float_matrix : Buffer.t -> float Umatrix.t -> unit;;
(** Same as above, for a C float matrix. *)

val send_fortran_float_matrix : Buffer.t -> float Umatrix.t -> unit;;
(** Same as above, for a Fortran float matrix. *)

(** {6 Input communication functions} *)

(** {7 Functions to receive communications} *)

(** Those functions receive a communication from an input buffer,
 and return the corresponding Caml value. *)

val receive_communication : Scanf.Scanning.scanbuf -> Communication.t;;
(** [receive_communication ib] reads a communication from the
 [!Scanf.Scanning.scanbuf] scanning buffer [ib]. *)

val receive_one_value : Scanf.Scanning.scanbuf -> Typed_value.t;;
(** [receive_one_value ib] reads a [!Communication.t] communication from the
 [!Scanf.Scanning.scanbuf] input buffer [ib]. The communication should be a
 [Phrase] with only one typed value, and the function returns this
 [!Typed_value.t] value. *)

val receive_values : Scanf.Scanning.scanbuf -> Typed_value.t array;;
(** [receive_values ib] reads a [Phrase] communication from the
 [!Scanf.Scanning.scanbuf] input buffer [ib] and returns the corresponding
 vector of typed values. *)

(** {7 Functions to receive typed values}
 
 Those functions receives a regular Caml value from an input buffer.
 
 More precisely, they return the regular Caml value corresponding to the
 {!Typed_value.t} value they read from their {Scanf.Scanning.scanbuf} input
 buffer argument. *)

val receive_string : Scanf.Scanning.scanbuf -> string;;
(** [receive_string ib] receives a typed value from the input buffer [ib]
 and converts it to a regular Caml [string] value. *)

val receive_int : Scanf.Scanning.scanbuf -> int;;
(** Same as above, for an integer. *)

val receive_int_couple : Scanf.Scanning.scanbuf -> int * int;;
(** Same as above, for an integer couple. *)

val receive_int_triple : Scanf.Scanning.scanbuf -> int * int * int;;
(** Same as above, for an integer triple. *)

val receive_float : Scanf.Scanning.scanbuf -> float;;
(** Same as above, for a float. *)

val receive_float_couple : Scanf.Scanning.scanbuf -> float * float;;
(** Same as above, for a float couple. *)

val receive_float_triple : Scanf.Scanning.scanbuf -> float * float * float;;
(** Same as above, for a float triple. *)

val receive_string_vector : Scanf.Scanning.scanbuf -> string array;;
(** Same as above, for a string array. *)

val receive_int_vector : Scanf.Scanning.scanbuf -> int array;;
(** Same as above, for an integer array. *)

val receive_int_couple_vector : Scanf.Scanning.scanbuf -> (int * int) array;;
(** Same as above, for an integer couple array. *)

val receive_int_triple_vector :
  Scanf.Scanning.scanbuf -> (int *  int * int) array
;;
(** Same as above, for an integer triple array. *)

val receive_float_vector : Scanf.Scanning.scanbuf -> float array;;
(** Same as above, for a float array. *)

val receive_float_couple_vector :
  Scanf.Scanning.scanbuf -> (float * float) array
;;
(** Same as above, for a float couple array. *)

val receive_float_triple_vector :
  Scanf.Scanning.scanbuf -> (float * float * float) array
;;
(** Same as above, for a float triple array. *)

val receive_c_int_matrix : Scanf.Scanning.scanbuf -> int Umatrix.t;;
(** Same as above, for a C integer matrix. *)

val receive_fortran_int_matrix : Scanf.Scanning.scanbuf -> int Umatrix.t;;
(** Same as above, for a Fortran integer matrix. *)

val receive_c_float_matrix : Scanf.Scanning.scanbuf -> float Umatrix.t;;
(** Same as above, for a C float matrix. *)

val receive_fortran_float_matrix : Scanf.Scanning.scanbuf -> float Umatrix.t;;
(** Same as above, for a Fortran float matrix. *)

(** {6 Foreman/Worker} *)

open Communication;;

type pf = Udebug.printing_facilities;;
type spf = Udebug.specialized_printing_facilities;;

type 'a injector = 'a -> values;;
type 'a projector = values -> 'a;;

type ('a, 'b) task_wrapper =
  {
   fname : Upervasives.function_name;
   tname : task_name;
   spf : spf;
   inj : 'a injector;
   proj : 'b projector;
  }
;;

exception Unbound_communication of t;;
exception Unbound_service_communication of (service_name * values);;

val make_task_wrapper :
  task_name -> (spf -> 'a injector) -> (spf -> 'b projector) ->
  pf -> Upervasives.function_name -> ('a, 'b) task_wrapper
;;

(** {7 Foreman} *)

val foreman_make_printing_facilities : unit -> pf;;

val foreman_wrap_task :
  Uunix.command_line -> Uunix.process_io_buffers option ref ->
  ('a, 'b) task_wrapper -> 'a -> 'b
;;

(** {7 Worker} *)

type ('a, 'b) task = spf -> 'a -> 'b;;

type welcome_action = Upervasives.gen_message;;
type farewell_action = Upervasives.gen_message;;

type task_computation =
  pf -> Uunix.process_io_buffers -> values -> unit
;;

type task_table = (task_name * task_computation) Ulist1.t;;

exception Unbound_task of task_name;;
exception Uninitialized_worker of task_name;;

val worker_wrap_task :
  Uunix.process_io_buffers ->
  ('b, 'a) task_wrapper -> ('a, 'b) task -> values -> unit
;;

val worker_make_printing_facilities : unit -> pf;;

val worker_main :
  welcome_action -> farewell_action -> task_table -> unit
;;

val worker_test_wrap_task :
  Uunix.process_io_buffers -> task_name -> 'a injector -> 'a -> unit
;;
